<* MAIN+ *>
MODULE BuildHash;

IMPORT
    In, Out,
    Perfect;

CONST
    numId = 256;

TYPE
    HashProc = PROCEDURE (VAR str: ARRAY OF CHAR; key: LONGINT): Perfect.HashInt;

VAR
    i, n, w1, w2, len, lmax: LONGINT;
    list: Perfect.HashList;
    d, c, p: Perfect.HashVector;
    H1, H2: HashProc;
    words: ARRAY numId, 32 OF CHAR;

PROCEDURE Length (VAR str: ARRAY OF CHAR): LONGINT;
VAR len: LONGINT;
BEGIN
    len := 0;
    WHILE str[len] # 0X DO INC(len) END;
    RETURN len;
END Length;

PROCEDURE Hash1 (VAR str: ARRAY OF CHAR; key: LONGINT): Perfect.HashInt;
VAR len: LONGINT;
BEGIN
    len := Length(str);
    IF key < 0 THEN
        INC(key, len);
        IF key < 0 THEN key := 0 END;
    ELSIF key >= len THEN
        key := len-1;
    END (*IF*);
    RETURN ORD(str[key]);
END Hash1;

PROCEDURE Hash2 (VAR str: ARRAY OF CHAR; key: LONGINT): Perfect.HashInt;
VAR len: LONGINT;
BEGIN
    len := Length(str);
    IF key < 0 THEN
        INC(key, len);
        IF key < 0 THEN key := 0 END;
    ELSIF key >= len THEN
        key := len-1;
    END (*IF*);
    RETURN SHORT(ORD(str[key]) + len);
END Hash2;

PROCEDURE MakePrimaryHash (Hash1, Hash2: HashProc): BOOLEAN;
VAR
    i, j: LONGINT;
    hash: ARRAY numId OF LONGINT;
BEGIN
    w1 := 0;
    WHILE w1 < lmax DO
        w2 := w1 + 1;
        WHILE w2 < lmax DO
            i := 0;
            LOOP
                IF i = n THEN
                    RETURN TRUE;
                END (*IF*);
                hash[i] := Hash1(words[i], w1) +
                           Hash2(words[i], w2)*65536;
                FOR j := 0 TO i-1 DO
                    IF hash[j] = hash[i] THEN
                        EXIT;
                    END (*IF*);
                END (*FOR*);
                INC(i);
            END (*LOOP*);
            INC(w2);
        END (*WHILE*);
        INC(w1);
    END (*WHILE*);

    RETURN FALSE;
END MakePrimaryHash;

PROCEDURE MakePrimaryHash2 (Hash1, Hash2: HashProc): BOOLEAN;
VAR
    i, j: LONGINT;
    hash: ARRAY numId OF LONGINT;
BEGIN
    w1 := -1;
    WHILE -w1 <= lmax DO
        w2 := w1 - 1;
        WHILE -w2 <= lmax DO
            i := 0;
            LOOP
                IF i = n THEN
                    RETURN TRUE;
                END (*IF*);
                hash[i] := Hash1(words[i], w1) +
                           Hash2(words[i], w2)*10000H;
                FOR j := 0 TO i-1 DO
                    IF hash[j] = hash[i] THEN
                        EXIT;
                    END (*IF*);
                END (*FOR*);
                INC(i);
            END (*LOOP*);
            DEC(w2);
        END (*WHILE*);
        DEC(w1);
    END (*WHILE*);

    RETURN FALSE;
END MakePrimaryHash2;

BEGIN
    In.Open; Out.Open;

    n := 0;
    LOOP
        i := -1;
        REPEAT
          INC(i);
          In.Char(words[n][i]);
        UNTIL ~In.Done OR (words[n][i] = 0AX);
        words[n][i] := 0X;
        IF ~In.Done THEN EXIT END;
        IF i # 0 THEN INC(n) END;
    END (*LOOP*);

    lmax := 0;
    FOR i := 0 TO n-1 DO
        len := Length(words[i]);
        IF lmax < len THEN lmax := len END;
    END (*FOR*);

    Out.String("PSEUDO-CODE:");
    Out.Ln();
    Out.String("word w is keyword if w = kw[f(w)]");
    Out.Ln();
    Out.String("f(w) = d[h1] + c[h1] mod p[h2]");
    Out.Ln();

    IF MakePrimaryHash(Hash1, Hash1) OR MakePrimaryHash2(Hash1, Hash1) THEN
        Out.String("h1 = w1 mod n");
        Out.Ln();
        Out.String("h2 = w2 mod m");
        H1 := Hash1;
        H2 := Hash1;
    ELSIF MakePrimaryHash(Hash1, Hash2) OR MakePrimaryHash2(Hash1, Hash2) THEN
        Out.String("h1 = w1 mod n");
        Out.Ln();
        Out.String("h2 = (w2 + length(w)) mod m");
        H1 := Hash1;
        H2 := Hash2;
    ELSIF MakePrimaryHash(Hash2, Hash1) OR MakePrimaryHash2(Hash2, Hash1) THEN
        Out.String("h1 = (w1 + length(w)) mod n");
        Out.Ln();
        Out.String("h2 = w2 mod m");
        H1 := Hash2;
        H2 := Hash1;
    ELSIF MakePrimaryHash(Hash2, Hash2) OR MakePrimaryHash2(Hash2, Hash2) THEN
        Out.String("h1 = (w1 + length(w)) mod n");
        Out.Ln();
        Out.String("h2 = (w2 + length(w)) mod m");
        H1 := Hash2;
        H2 := Hash2;
    ELSE
        Out.String("Can't select a primary hash :(");
        HALT;
    END (*IF*);

    Out.Ln();
    Out.String("w1 = if length(w) > ");
    Out.Int(w1, 0);
    Out.String(" then w[");
    Out.Int(w1, 0);
    Out.String("] else w[length(w)-1]");
    Out.Ln();
    Out.String("w2 = if length(w) > ");
    Out.Int(w2, 0);
    Out.String(" then w[");
    Out.Int(w2, 0);
    Out.String("] else w[length(w)-1]");
    Out.Ln();

    NEW(list, n);

    FOR i := 0 TO n-1 DO
        list[i].index := i;
        list[i].h1 := H1(words[i], w1);
        list[i].h2 := H2(words[i], w2);
    END (*FOR*);

    IF Perfect.MakeHash(list, d, c, p) THEN
        Out.String("n = size(c) = size(d) = ");
        Out.Int(LEN(c^), 0);
        Out.Ln();
        Out.String("m = size(p) = ");
        Out.Int(LEN(p^), 0);
        Out.Ln();

        Out.String('kw = {');
        FOR i := 0 TO n-1 DO
            Out.String('"');
            Out.String(words[list[i].index]);
            Out.String('",');
            Out.Ln();
        END (*FOR*);
        Out.String('}');
        Out.Ln();
    
        Out.String("d = {");
        FOR i := 0 TO LEN(d^)-2 DO
            Out.Int(d[i], 0);
            Out.String(",");
        END (*FOR*);
        Out.Int(d[LEN(d^)-1], 0);
        Out.String("}");
        Out.Ln();

        Out.String("c = {");
        FOR i := 0 TO LEN(c^)-2 DO
            Out.Int(c[i], 0);
            Out.String(",");
        END (*FOR*);
        Out.Int(c[LEN(c^)-1], 0);
        Out.String("}");
        Out.Ln();

        Out.String("p = {");
        FOR i := 0 TO LEN(p^)-2 DO
            Out.Int(p[i], 0);
            Out.String(",");
        END (*FOR*);
        Out.Int(p[LEN(p^)-1], 0);
        Out.String("}");
        Out.Ln();
    ELSE
        Out.String("Can't build a perfect hash function :(");
    END (*IF*);
END BuildHash.
